/**********************************************************************************
   Copyright(c) Catalyst Semiconductor, Inc., 2006. All rights reserved.

This code is provided as is without warranty or guarantee of any kind as its accuracy, 
completeness, operability, fitness for particular purpose, or any other warranty.

   File        :  IICmaster.v
   Description :  Verilog model for IIC protocol - master side
		  for testing Verilog model for CAT24XXXX I2C serial EEPROMs

   Version     :  Rev1.01 
   Release     :  Sep, 2006
**********************************************************************************/

module IICmaster(SDA, SCL, WPPAD,A0PAD,A1PAD,A2PAD,A0,A1,A2,prefi1,prefi2,prefi3,prefi4);
//
//	Master for i2c protocol 
//
output 		SCL,WPPAD;	
reg             SCL,WPPAD;

inout 		SDA;				

input A0PAD,A1PAD,A2PAD,A0,A1,A2,prefi1,prefi2,prefi3,prefi4;

//
`include "MemType.v"
//------------------
//
`define B0			1'b0
`define B1			1'b1
`define Y			1'b0
`define N			1'b1

`define afterNoAckStart		2'b00
`define afterStopStart		2'b01
`define afterNoStopStart	2'b10

`define idle 			5'b00000
`define sendingAddrInfo 	5'b00001
`define sendingAdd 		5'b00010
`define sendingData		5'b00111
`define sendingNextData	5'b01000
`define sendingAck		5'b01001
`define sendingStart		5'b01010
`define sendingStop		5'b01011
`define receivingData		5'b01100
`define waitingAck		5'b01001

`define write 			1'b0
`define read 			1'b1

`define byteWrite 		5'b00000
`define pageWrite 		5'b00001
`define immedAddrRead 		5'b00010
`define selectiveRead 		5'b00011
`define sequentialImmedRead 	5'b00100
`define sequentialSelectRead 	5'b00101

`define nothingSpecial		3'b000
`define startToSend		3'b001
`define ackReceived 		3'b010
`define noAckReceived 		3'b011
`define stopToSend		3'b100

`define OUT			2'b00
`define HIGH			2'b01

//===============================================================================

parameter sepLength=8;
parameter frequency = 100000; //(Hz)
parameter delta = 2;

//===============================================================================		

reg[2:0] addrDev, realAddrDev;
reg[7:0] addrInfoByteR;
reg[7:0] addrInfoByteW; 
reg[3:0] headerFirstHalf,reqHeaderFirstHalf;	// really sent & required   first half header
reg[8*10:1] chipName;

reg [7:0] bufWriteData[0:4095];		
reg [7:0] bufReadData[0:4095];
reg [7:0] MpageReg[0:8];	
reg [7:0] linestr[0:79];
reg [7:0] MtempReg,MdataReg,manRegB; 

reg [1:0] startType;
reg[4:0] Mstate,prevMstate,MnextState,MafterAckState;
reg [1:0] MbusState; // State machine controller
reg [2:0] Mcondition;			 
reg[1:0] Moperation = `write;		// Operation (Read/Write) 
reg[4:0] MoperationType = `pageWrite;	// Operation type (certain kind of read or write)

reg[15:0] addr, manAddr, manAddr1;
reg[22:0] manRegW;
reg[15:0] crtMemAddr = 16'h0000;
reg[15:0] realCrtMemAddr = 16'h0000;
reg[7:0] dataByte = 8'b10101010;
reg[3:0] genDataCell,pattern,patternIni;

reg currentOperation;
reg MregToBus =1'bz;
reg alreadyInfoDisplay, LongOperation;
reg Polling = `N;

integer fMAX, tBUF, tSU_STA, tHD_STA, tSU_STO, tSU_DAT, tHD_DAT, tDH, tAA, tWR;
integer tHALF, tMID;
integer MbitCounter,refMbitCounter, pageCounter, seqCounter, recDataCounter, seqLength, lSeq;
integer semFirst, semForever, selectiveReadPhase,deltaTime;
integer TestPhase,addrBytes,addrBytesCounter,lgDisp, notRec, notRecCount;
integer slaveAddrBits;
integer allMemLength, lPage, Length, savLength, initialDelay, protectType;
integer i, j;
integer man0, man;

//===============================================================================	

event reinitClockForStart, reinitReady;

//===============================================================================
	
tri1 SDA;
assign SDA = (MbusState == `OUT) ? MregToBus : 1'bz;

//
//#################################### TASKS ###########################################
//
task createIncremFillPattern;
	begin
	genDataCell = 4'b0000;
	patternIni = pattern;
	for(i=0;i<lSeq;i=i+1)
		begin
		// pattern = i/lPage;
		bufWriteData[i] = {pattern,genDataCell};
		genDataCell = genDataCell+1;
		pattern = patternIni + (i+1)/16;
		end
	end
endtask

task displayData;
	begin
	// $display("content of memory block after current operation:");
	if(seqLength == 1)
		$display("----------");
	else
		$display("---------------------------------------------------------------------");
	for(i=0;i<seqLength+1;i=i+1)
		begin
		j=i%16;
		if(j == 0 && i>0)
			begin
			manAddr = addr+i-16;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8],linestr[9],linestr[10],linestr[11],linestr[12],linestr[13],linestr[14],linestr[15]);
			linestr[j] = bufWriteData[i];
			end
		else
			linestr[j] = bufWriteData[i];
		end
	if(j>0)
		begin
		case(j)
		1:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h",manAddr,linestr[0]);
			end
		2:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h",manAddr,linestr[0],linestr[1]);
			end
		3:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2]);
			end
		4:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3]);
			end
		5:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4]);
			end
		6:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5]);
			end
		7:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6]);
			end
		8:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7]);
			end
		9:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8]);
			end
		10:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8],linestr[9]);
			end
		11:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8],linestr[9],linestr[10]);
			end
		12:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8],linestr[9],linestr[10],linestr[11]);
			end
		13:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8],linestr[9],linestr[10],linestr[11],linestr[12]);
			end
		14:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8],linestr[9],linestr[10],linestr[11],linestr[12],linestr[13]);
			end
		15:
			begin
			manAddr = addr+i-j-1;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8],linestr[9],linestr[10],linestr[11],linestr[12],linestr[13],linestr[14]);
			end

		default:
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8],linestr[9],linestr[10],linestr[11],linestr[12],linestr[13],linestr[14],linestr[15]);				
		endcase
		end
	if(seqLength == 1)
		$display("----------");
	else

		$display("---------------------------------------------------------------------");
	// $display("current address: %h    time: %9t",crtMemAddr,$time);
	end
endtask

task initBufReadData;
	begin
	for(i=0;i<4096;i=i+1)
		bufReadData[i] = 8'hff;	
	end
endtask

task displayRecData;
	begin
	//$display("Master: %d bytes of received data:",recDataCounter);
	if(recDataCounter == 1)
		begin
		$display("----------");
		lgDisp = 1;
		end
	else
		begin
		$display("---------------------------------------------------------------------");
		lgDisp = recDataCounter;
		if(lgDisp > allMemLength)
			lgDisp = allMemLength;
		end
//
//	for(i=0;i<recDataCounter+1;i=i+1)

	for(i=0;i<lgDisp+1;i=i+1)
		begin
		j=i%16;
		if(j == 0 && i>0)
			begin
			manAddr = addr+i-16;
			$display("%h:  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h  %h",manAddr,linestr[0],linestr[1],linestr[2],linestr[3],linestr[4],linestr[5],linestr[6],linestr[7],linestr[8],linestr[9],linestr[10],linestr[11],linestr[12],linestr[13],linestr[14],linestr[15]);
			linestr[j] = bufReadData[i];
			end
		else
			linestr[j] = bufReadData[i];
		end
	if(j>0)
		begin
		case(j)
		1:
			begin
			manAddr = addr+i-j-1;
			$display("%h:   %h",manAddr,linestr[0]);
			end	
		default:
			i=i;
		endcase
		end
	if(recDataCounter == 1)
		$display("----------");
	else

		$display("---------------------------------------------------------------------");
	manRegW =  $time;
	// $display("current address: %h,  time: %d",crtMemAddr,manRegW);
	end
endtask

task CheckAddr;
	begin
		manAddr = allMemLength-1;
		if(addr[15:0] > (allMemLength-1))
			begin
			$display;
			$display(" *** MASTER - WARNING: ");
			$display(" *** Specified address (%h) higher than maximum allowed (%h) !",addr,manAddr);
			addr = addr%allMemLength;
			$display(" *** Address %h will be used instead !", addr);
			$display;
			end

	end
endtask

task SetHeader;
	begin
		`ifdef cat24c01
			addrInfoByteW = {headerFirstHalf,addrDev[2:0],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2:0],`B1};
		`endif

		`ifdef cat24c02
			addrInfoByteW = {headerFirstHalf,addrDev[2:0],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2:0],`B1};
		`endif

		`ifdef cat24c03
			addrInfoByteW = {headerFirstHalf,addrDev[2:0],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2:0],`B1};
		`endif

		`ifdef cat24c04
			addrInfoByteW = {headerFirstHalf,addrDev[2:1],addr[8],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2:1],addr[8],`B1};
		`endif

		`ifdef cat24c05
			addrInfoByteW = {headerFirstHalf,addrDev[2:1],addr[8],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2:1],addr[8],`B1};
		`endif
		
		`ifdef cat24c08
			addrInfoByteW = {headerFirstHalf,addrDev[2],addr[9:8],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2],addr[9:8],`B1};
		`endif

		`ifdef cat24c16
			addrInfoByteW = {headerFirstHalf,addr[10:8],`B0}; 
			addrInfoByteR = {headerFirstHalf,addr[10:8],`B1};
		`endif

		`ifdef cat24c164
			addrInfoByteW = {headerFirstHalf,addr[10:8],`B0}; 
			addrInfoByteR = {headerFirstHalf,addr[10:8],`B1};
		`endif

		`ifdef cat24c32
			addrInfoByteW = {headerFirstHalf,addrDev[2:0],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2:0],`B1};
		`endif

		`ifdef cat24c64
			addrInfoByteW = {headerFirstHalf,addrDev[2:0],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2:0],`B1};
		`endif

		`ifdef cat24c128
			addrInfoByteW = {headerFirstHalf,addrDev[2:0],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2:0],`B1};
		`endif

		`ifdef cat24c256
			addrInfoByteW = {headerFirstHalf,addrDev[2:0],`B0}; 
			addrInfoByteR = {headerFirstHalf,addrDev[2:0],`B1};
		`endif

	end
endtask

//################################################# INITIAL #############################
//
//	initial
//
initial

	begin

	tHALF = 1000000000/(2*frequency); 
	tMID = (3/8)*tHALF;
	WPPAD = 0;
	MregToBus = 1'bz;

	initialDelay = 2000;

	#1 manRegB = 0;	
	
	realAddrDev = {A2PAD,A1PAD,A0PAD};
	addrDev = {A2,A1,A0};

	headerFirstHalf = {prefi1,prefi2,prefi3,prefi4};
	addrInfoByteW = {headerFirstHalf,3'b000,`B0}; 
	addrInfoByteR = {headerFirstHalf,3'b000,`B1};

	`include "MemParameters.v"
	//------------------------

	#1 case(slaveAddrBits)
		3:
			$display("*     Used device address:   %b (A2,A1,A0)",addrDev[2:0]);
		2:
			$display("*     Used device address:   %b (A2,A1)",addrDev[2:1]);
		1:
			$display("*     Used device address:   %b (A2)",addrDev[2]);
	endcase

	$display("*===============================================================");
	$display;

	man0 = frequency;
	man = 2*tHALF;
	$display("* Tclock  = %d ns (Frequency: %d Hz)",man,man0);
	$display("*");
	#0 if(man0 > 1000*fMAX)
		begin
		//$display;
		$display("*===============================================================================");
		$display("*\t WARNING: USED CLOCK FREQUENCY IS HIGHER THAN THE ALLOWED MAXIMUM ONE !");
		$display("*\t PLEASE, DO MODIFY THE CLOCK FREQUENCY.");
		$display("*===============================================================================");
		$display;
		$stop;
		end

	TestPhase = 1;
	LongOperation = `N;
	alreadyInfoDisplay = `N;
	pattern = 0;
//
//	First step in test
//
	`include "Input0.v"
	//----------------

	CheckAddr;
	SetHeader;

	manRegB = TestPhase;
	manRegW = $time;	
	#0 if(WPPAD == 1)
		$display("PROTECTED write,  test #%d,  time: %d",manRegB,manRegW); 
	else
		$display("UNPROTECTED write,  test #%d,  time: %d",manRegB,manRegW); 
//
	TestPhase = 2;
	Mstate = `idle;
	Mcondition = `nothingSpecial;
	MnextState = `sendingStart;
	MregToBus = 1'bz;
	MbusState = `OUT;
	currentOperation = `write;
	selectiveReadPhase = 1;
	startType = `afterNoAckStart;
	MbitCounter = 0;
	pageCounter = lSeq;
	seqCounter = 0;
	SCL=1'b1;
	semFirst = 0;
	semForever = 0;
	#(initialDelay+tHALF/5) SCL=`B1;
	#tHD_STA
	while(semForever == 0)
		#(tHALF) if(semForever == 0) 
				SCL = ~SCL;
			// just a first double tHIGH to allow a good START Mcondition
			// from now until STOP cond, clock period will be the normal one
	 ->reinitReady;
	end
//
//################################################# ALWAYS #########################
//
//	change SCL for (STOP) START sending
//
always @(reinitClockForStart)
	begin
	SCL = 1'b1;
	Length = sepLength;
	if(startType == `afterStopStart)
		begin
		deltaTime = tHALF*sepLength;
		end
	 else
		begin
		if(startType == `afterNoStopStart)
			begin
			savLength =  sepLength;
			deltaTime = tHALF;			
			Length = 1;
			end
		else
			deltaTime = tHALF*sepLength;
		end
	#deltaTime SCL = 1'b1;
	semForever = 0;
	#tHD_STA 
	while(semForever == 0)
		begin
		#tHALF if(semForever == 0) 
			SCL = ~SCL;
			// just a first double tHIGH to allow a good START Mcondition
			// from now until STOP cond, clock period will be the normal one
		end
	->reinitReady;
	end				
//
//	STOP sending
//
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
always @(posedge SCL) 
	begin
	//$display("___ Master: SCL + (send Stop) ",$time);
	//$display("___ Mcondition = %b MbitCounter = %d ", Mcondition,MbitCounter,$time);

	#tSU_STO 	if(Mcondition == `stopToSend && MbitCounter == 0) 
		begin

		$display("SENDING STOP ",$time);

		MbusState = `OUT;
		MregToBus = 1'bz;
		Mcondition = `nothingSpecial;
		#tMID MbusState = MbusState;
		currentOperation = `write;
		seqCounter = 0;
		#(tHALF-tSU_STO-tMID-tMID/8) alreadyInfoDisplay = `N;
		if(LongOperation == `Y)
			begin
			TestPhase = TestPhase - 1;
			LongOperation = `N;
			end
		else
			begin
			$display("current address: %h,  time: %d",realCrtMemAddr,manRegW);
			$display;
			$display;
			$display;
			selectiveReadPhase = 1;
			manRegB = TestPhase;
			manRegW = $time;

			case(TestPhase)
				`include "Input.v"
				default:
				$stop;
			endcase
			if(WPPAD == 1)
				$display("PROTECTED write,  test #%d,  time: %d",manRegB,manRegW); 
			else
				$display("UNPROTECTED write,  test #%d,  time: %d",manRegB,manRegW); 
			end

		CheckAddr;
		TestPhase = TestPhase + 1;
		pageCounter = lSeq;	
		end
	end


///////////////////////////////////////////////////////////
//
//	State machine for RECEIVE type Master activity 
//
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
always @(posedge SCL && Mstate != `idle )
	begin
	//$display("___ Master: SCL + (receive) ",$time);
	if (currentOperation == `write && MbitCounter == 9)
		begin
		currentOperation = currentOperation;
		end
	else
		begin
		//
		//	State machine for RECEIVE type Master activity 
		//
		case(Mstate)

		`receivingData:
			begin
			// prevMstate = `receivingData;
			case(MbitCounter)
			9: 
				begin
				if(refMbitCounter == 0)
					begin
					MbitCounter = 1; 
					MtempReg[7:0] = 8'b00000000;
					MtempReg[0] = SDA;
					//$display("Master: receiving data     %b",MtempReg[0]);
					end
				end

			// cases 1,2,3,4,5,6,7,8
			default:
				begin
				MtempReg = MtempReg << 1;
				MtempReg[0] = SDA;
				// $display("Master: receiving data     %b %d ",MtempReg[0],MbitCounter);
				end
			endcase

			if(MbitCounter == 8)
				begin
				#(tHALF+tDH) MbusState = `OUT;
				MregToBus = `B0;	
				// ### #(tHALF-tDH) MbitCounter = MbitCounter+1;
				#(tHALF-tDH-1) MbitCounter = MbitCounter+1;	
				end
			else
				begin
				// ### #(tHALF*2) if(MbitCounter == 9 && refMbitCounter == 9)
				#(tHALF*2-1) if(MbitCounter == 9 && refMbitCounter == 9)
					begin
					if(MoperationType == `immedAddrRead || MoperationType == `selectiveRead || recDataCounter == lSeq)
						MbitCounter = 0;
					else
						MbitCounter = 1;
					end
				else
					begin
					MbitCounter = MbitCounter+1;
					refMbitCounter = 9;
					end
				end
			end

		`sendingStop:
			begin
			$display("Prepare SEND STOP ",$time);
			#0 Mcondition = `stopToSend;
			MregToBus = `B0;
			MbusState = `OUT;
			pageCounter = lSeq;
			MnextState = `sendingStart;
			end
		endcase
	end
end



//--------------------------------------------------------------------------------------------
always @(negedge SCL && Mstate != `idle )
	//
	//	State machine for SEND type Master activity 
	// 
	begin
	//
	//	state transition
	//
	//$display("___ Master: SCL - (send) ",$time);
	if(Mstate == `receivingData && MbitCounter == 8 )
		begin
		#(tHALF-tSU_DAT) if(MoperationType == `immedAddrRead || MoperationType == `selectiveRead || seqCounter == lSeq-1)
			begin
			bufReadData[recDataCounter] = MtempReg[7:0];
			recDataCounter = recDataCounter + 1;
			crtMemAddr = crtMemAddr + 1;	//(inc.adr.in pag)
			realCrtMemAddr = crtMemAddr;
			displayRecData;
			//$display("Master: sending NO ACK ",$time);
			MregToBus = 1'bz;							
			#tHALF startType = `afterNoAckStart;
			#0 Mcondition = `stopToSend;
			// ### #(tSU_DAT+tDH) 
			#(tSU_DAT-1)
			Mstate = `sendingStop;
 			MnextState = `sendingStop;
			end
		else
			begin
			//$display("Master: sending ACK ",$time);
			//$display("Master: received data %d : %h",seqCounter+1,MtempReg[7:0]);
			crtMemAddr = crtMemAddr + 1;	//(inc.adr.in pag)
			realCrtMemAddr = crtMemAddr;
			bufReadData[recDataCounter] = MtempReg[7:0];
			recDataCounter = recDataCounter + 1;
			#(tSU_DAT+tHALF+tDH) MregToBus = 1'bz;
			#(tMID/2) MbusState = `HIGH;
			seqCounter = seqCounter + 1;
			MnextState = `receivingData;
			MdataReg[7:0] = MtempReg[7:0];
			MpageReg[pageCounter] = MtempReg[7:0];
			MdataReg[7:0] = MtempReg[7:0];
			#0 if(pageCounter >= lPage) pageCounter = 0;
			end
		end
	prevMstate = Mstate;	
	Mstate = MnextState;
	case(Mstate)
		`sendingStart:
			begin
			//$display(" START    ",$time);
				//
				//	master generates START condition
				//
				if(semFirst == 1)
					begin
					semForever = 1;
					SCL = 1;
					@(reinitReady)
						->reinitClockForStart; 
					#tSU_STA Mcondition = `startToSend;
					end
				else
					begin
					#tSU_STA	Mcondition = `startToSend;
					end


				if(semFirst == 0)
					deltaTime = 0;
				else
					deltaTime = tHALF;

				#deltaTime 
				
				if(semFirst == 0)
					begin
					#(initialDelay+tMID) MregToBus = `B0;
					end
				else
					begin
					#((Length-1)*tHALF) MregToBus = `B0;
					end

				MbusState = `OUT;

				$display("Master: sending START at time ",$time);
			MnextState = `sendingAddrInfo;
			semFirst = 1;
			end

		`sendingAddrInfo: 
			begin
			Length =  savLength;
			Polling = `Y;
			currentOperation = `write;
			#tHD_DAT MregToBus = `B0;
			if(MbitCounter == 0)
				begin
				if(Moperation == `write || (Moperation == `read && (MoperationType == `selectiveRead || MoperationType == `sequentialSelectRead) && selectiveReadPhase == 1) )
					MtempReg = addrInfoByteW;
				else
					MtempReg = addrInfoByteR;
			//	$display("Master: sent addrInfo:         %b",MtempReg);
				end
			#(tHALF-tSU_DAT-tHD_DAT)  if(MbitCounter < 9)
				begin 
				MbitCounter = MbitCounter + 1;
				refMbitCounter = MbitCounter;
				end
			else
				begin
				MbitCounter = 1;
				refMbitCounter = 9;
				end
			if(refMbitCounter <9)
				MbusState = `OUT;

			if(refMbitCounter == 9) 
				begin
				case(Moperation)
				`write:	
					begin
					if(MoperationType == `byteWrite)
						begin
						if(alreadyInfoDisplay == `N)
							begin
							$display("BYTE WRITE");
							$display("address: %h",addr);
							alreadyInfoDisplay = `Y;
							end
						end
					else
						begin
						createIncremFillPattern;
						if(alreadyInfoDisplay == `N)
							begin
							$display("PAGE WRITE");
							manRegB = lSeq;
							manAddr = lSeq;
							$display("address: %h, length: %d (%h h)",addr,manRegB,manAddr);
							seqLength = lSeq;

							displayData;
							alreadyInfoDisplay = `Y;
							end
						end
					crtMemAddr = addr;
					realCrtMemAddr = crtMemAddr;
					MafterAckState = `sendingAdd;
					addrBytesCounter = addrBytes;
					end

				`read:
					begin
					addrBytesCounter = addrBytes;
					recDataCounter = 0;
					initBufReadData;
					case(MoperationType)
					`immedAddrRead:
						begin
						if(alreadyInfoDisplay == `N)
							begin
							$display("IMMEDIATE ADDRESS READ");
							addr = realCrtMemAddr;
							alreadyInfoDisplay = `Y;
							end
						MafterAckState = `receivingData; 
						currentOperation = `read;
						end
					`sequentialImmedRead: 
						begin
						if(alreadyInfoDisplay == `N)
							begin
							$display("SEQUENTIAL IMMEDIATE ADDRESS READ");
							addr = realCrtMemAddr;
							manAddr = lSeq;
							$display("length: %d (%h h)",manAddr,manAddr);
							alreadyInfoDisplay = `Y;
							end
						MafterAckState = `receivingData; 
						currentOperation = `read;
						end
					`selectiveRead:
						begin
						if(alreadyInfoDisplay == `N && selectiveReadPhase == 1)
							begin
							$display("SELECTIVE READ");
							$display("address: %h",addr);
							alreadyInfoDisplay = `Y;
							end
						crtMemAddr = addr;
						realCrtMemAddr = crtMemAddr;
						if(selectiveReadPhase == 1)
							MafterAckState = `sendingAdd; 
						else
							begin
							MafterAckState = `receivingData;
							currentOperation = `read;
							end
						end
					`sequentialSelectRead: 
						begin
						if(alreadyInfoDisplay == `N  && selectiveReadPhase == 1)
							begin
							$display("SEQUENTIAL SELECTIVE READ");
							manRegB = lSeq;
							manAddr = lSeq;
							$display("address: %h, length: %d (%h h)",addr,manAddr,manAddr);
							alreadyInfoDisplay = `Y;
							end
						crtMemAddr = addr;
						realCrtMemAddr = crtMemAddr;
						if(selectiveReadPhase == 1)
							MafterAckState = `sendingAdd; 
						else
							begin
							MafterAckState = `receivingData;
							currentOperation = `read;
							end
						end
					endcase
					end
				endcase
				MnextState = MafterAckState;
				refMbitCounter = 0;
				end
			else
				begin
				MregToBus = MtempReg[7];
				#0 if(MregToBus==1'b1)
					MregToBus = 1'bz;
				//	$display("      // Master: sending addrInfo     %b",MregToBus);
				#0 MtempReg = MtempReg << 1;
				end
			end

		`sendingAdd:
			begin
			// prevMstate = `sendingAdd;
			Polling = `N;
			MbusState = `OUT;
			#tHD_DAT MregToBus = `B0;					
			if(MbitCounter == 9) 
				begin
				if(addrBytesCounter == 2) 
					begin
					MtempReg = addr[15:8];
					//$display("Master: sent addr[15:8]:             %h",MtempReg);
					MnextState = `sendingAdd;
					#(tHALF-tHD_DAT-tSU_DAT) MregToBus = MtempReg[7];
					#0 if(MregToBus==1'b1)
						MregToBus = 1'bz;
					#0 MtempReg = MtempReg << 1;
					#tSU_DAT MbitCounter = 1;
					end
				else
					begin
					MtempReg = addr[7:0];
					if(Moperation == `write)
						begin
						//$display("Master: sent addr[ 7:0]:             %h",MtempReg);
						MafterAckState = `sendingData;
						refMbitCounter = 0;
						#(tHALF-tHD_DAT-tSU_DAT) MregToBus = MtempReg[7];
						#0 if(MregToBus==1'b1)
							MregToBus = 1'bz;
						#0 MtempReg = MtempReg << 1;
						#tSU_DAT MbitCounter = 1;
						end
					else
						begin
						if(addrBytesCounter == 1) 
							begin
							// $display("Master: sent addr[ 7:0]:             %h",MtempReg);
							selectiveReadPhase = 2;
							#(tHALF-tHD_DAT-tSU_DAT) MregToBus = MtempReg[7];
							#0 if(MregToBus==1'b1)
								MregToBus = 1'bz;
							#0 MtempReg = MtempReg << 1;
							#tSU_DAT MbitCounter = 1;
							addrBytesCounter = addrBytesCounter-1;
							end
						else
							begin
							if(Moperation == `write)
								begin
								Mcondition = `stopToSend;
								Mstate = `stopToSend;
								MafterAckState = `sendingStop;
								startType = `afterNoAckStart;
								MnextState = `sendingStart;
								#(tHALF-tHD_DAT-delta) MbitCounter = 0;
								LongOperation = `Y;
								end
							else
								/* No Stop for Selective Read */
								begin
								#(tHALF/2) MbusState = `OUT;
								MregToBus = 1'bz;
								Mcondition = `stopToSend;
								Mstate = `stopToSend;
								MafterAckState = `sendingStop;
								startType = `afterNoStopStart;
								MnextState = `sendingStart;
								#(tHALF/2-tHD_DAT-delta) MbitCounter = 0;
								LongOperation = `Y;
								end

							end
						end
					end
				end
			else
				begin
				#(tHALF-tHD_DAT-tSU_DAT) MregToBus = MtempReg[7];
				#0 if(MregToBus==1'b1)
					MregToBus = 1'bz;
				// $display("Master: sending address     %b",MregToBus );
				#0 MtempReg = MtempReg << 1;
				if(MbitCounter == 8)
					begin
					addrBytesCounter = addrBytesCounter-1;
					MnextState = MafterAckState;
					MbusState = `HIGH; 
					end
				#tSU_DAT MbitCounter = MbitCounter + 1;
				end
			end

		`sendingData:
			begin
			Polling = `N;
			#tHD_DAT MregToBus = `B0;
			//
			//	first MbitCounter = 9
			//
			if(MbitCounter == 9 && refMbitCounter == 0)
				begin
				MnextState = MafterAckState;
				case(MoperationType)
					`byteWrite:
						begin
						MtempReg = dataByte;
						end
					`pageWrite:
						begin
						MtempReg = bufWriteData[lSeq-pageCounter];
						end
				endcase
				// $display("Master: sent data %d: %h",lSeq-pageCounter+1,MtempReg);
				crtMemAddr[`pgExpMinusOne:0] = crtMemAddr[`pgExpMinusOne:0] + 1;//(inc.adr.in pag)
				realCrtMemAddr = realCrtMemAddr + 1;
				/*********** NOT YET TREATED !!!
				if(realCrtMemAddr != crtMemAddr)	// end of page
					$display("==== %d  ==== ",pageCounter);
				********************/
				end	
			#(tHALF-tHD_DAT-tSU_DAT) MregToBus = MtempReg[7];
			#0 if(MregToBus==1'b1)
				MregToBus = 1'bz;

			#0	MtempReg = MtempReg << 1;
			#tSU_DAT if(MbitCounter == 9 && refMbitCounter == 0 )
				begin
				MbitCounter = 1; 
				end
			else
				begin
				MbitCounter = MbitCounter+1; 
				end
			if(MbitCounter == 8 )
				begin
				refMbitCounter = 9;
				end
			//
			//	second MbitCounter = 9
			//
			if(MbitCounter == 9 && refMbitCounter == 9)
				begin
				case(MoperationType)
					`byteWrite:
						begin
						Mcondition = `stopToSend;
						MafterAckState = `sendingStop;
						Mstate = `sendingStop;
						startType = `afterStopStart;
						MnextState = `sendingStart;
						$display("----------");
						$display("%h:   %h",addr,dataByte);
						#(tHALF) $display("----------");
						#(tHALF-tSU_DAT) MbitCounter = 0;
						end
					`pageWrite:
						begin
						if(pageCounter > 1)
							begin
							MafterAckState = `sendingData;
							refMbitCounter = 0;
							end
						else
							begin
							Mcondition = `stopToSend;
							MafterAckState = `sendingStop;
							startType = `afterStopStart;
							MnextState = `sendingStop;
							Mstate = MnextState;
							end

						if(Mcondition == `noAckReceived)
							begin
							Mcondition = `stopToSend;
							MafterAckState = `sendingStop;
							startType = `afterStopStart;
							MnextState = `sendingStart;
							pageCounter = lSeq;
							end
						#(2*tHALF-tSU_DAT) if(Mcondition == `noAckReceived || pageCounter == 1)
							MbitCounter = 0;
						else
							begin
							MbitCounter = 1;
							if(MbitCounter == 1)
								pageCounter = pageCounter - 1;
							MtempReg = bufWriteData[lSeq-pageCounter];

							crtMemAddr[`pgExpMinusOne:0] = crtMemAddr[`pgExpMinusOne:0] + 1;//(inc.adr.in pag)
							realCrtMemAddr = realCrtMemAddr + 1;


							if(realCrtMemAddr != crtMemAddr && pageCounter > 1)	// end of page
					
								realCrtMemAddr = crtMemAddr;
							MregToBus = MtempReg[7];
							#0 if(MregToBus==1'b1)
								MregToBus = 1'bz;
							#0 MbusState = `OUT;
							#0 MtempReg = MtempReg << 1;
							end
						end
				endcase
				end
			end
	endcase 
	end	
  
//---------------------------------------------------------------------------------------------
//	set conditions for ACK detection
//
always @(negedge SCL && Mstate != `idle )
	if (currentOperation == `write && MbitCounter == 8)
			begin
			//$display("___ Master: SCL - (set det Ack) ",$time);
			MbusState = `OUT;
			#0 MregToBus = 0;	
			#tMID MbusState = `HIGH;
			end


//---------------------------------------------------------------------------------------------
//	ACK detection  & reset conditions for ACK detection
//
 always @(posedge SCL && Mstate != `idle )
	begin
	//$display("___ Master: SCL + (det ACK) ",$time);
	#0 if ((currentOperation == `write || (currentOperation == `read && prevMstate == `sendingAddrInfo )) && MbitCounter == 9)

		begin
		if(SDA == `B0)
			begin
			//$display("=== Master: ACK received ",$time);
			MnextState = MafterAckState;
			manRegW = $time;
			//$display("current address: %h,  time: %d",crtMemAddr,manRegW);
			if ((MoperationType == `selectiveRead || MoperationType == `sequentialSelectRead ) && Mstate == `sendingStart) 
				MregToBus = 1'bz;
			end
		else
			begin
			$display("Master:   NO ACK received at time: ",$time);
			crtMemAddr[`pgExpMinusOne:0] = crtMemAddr[`pgExpMinusOne:0] - 1;//(dec.adr.in pag)
			realCrtMemAddr = realCrtMemAddr - 1;
			manRegW = $time;
			// $display("current address: %h,  time: %d",crtMemAddr,manRegW);
			if (prevMstate == `sendingAddrInfo)	
				begin
				notRec = 0;
				notRecCount = 0;
				if(notRec == 1)
					begin
					$display("      TEST ENDED");
					$stop;
					end
				else
					begin
					$display("Master:   ADDRESSED DEVICE NOT RESPONDING !",$time);
					notRecCount = notRecCount+1;
					#tHALF MbitCounter= 0;
					if(notRecCount > 10)
						begin
						$display("      TEST ENDED !");
						$stop;
						end
					end
				end
 			Mcondition = `noAckReceived;
			Polling = `N;
 			Mcondition = `noAckReceived;
			startType = `afterStopStart;
			MregToBus = 1'bz;
			MnextState = `sendingStop;
			Mstate = `sendingStop;
			Mcondition = `sendingStop;
			end
		 #(tHALF-delta) 	
		if(currentOperation == `write)
			begin
			MbusState = `OUT;
			MregToBus = `B0;
			end
		end
	end

endmodule

